@@ -3,10 +3,12 @@ from dataclasses import dataclass
 from distutils.version import LooseVersion
 import logging
 from pathlib import Path
+import shutil
 import sys
 import tempfile
 from typing import Tuple, Optional, Iterable, Set
 
+import appdirs  # type: ignore
 import click
 
 from .local import (
@@ -23,6 +25,7 @@ from .remote import (
 @dataclass
 class MinetestUserData:
     userpath: Path
+    cache: Path
 
 
 class ApplicationError(click.ClickException):
@@ -88,8 +91,19 @@ def dedup_packages(package_strs: Tuple[str, ...]) -> Set[str]:
     default=Path.cwd(),
     help="Path to the Minetest user data directory, usually named '.minetest'. Defaults to the current directory.",
 )
+@click.option(
+    "--clear-cache",
+    is_flag=True,
+    help="Force the package metadata cache to be cleared.",
+)
 @click.pass_context
-def main(ctx: click.Context, verbosity: int, quiet: bool, userpath: Optional[Path]):
+def main(
+    ctx: click.Context,
+    verbosity: int,
+    quiet: bool,
+    userpath: Optional[Path],
+    clear_cache: bool,
+):
     """A command-line package manager for Minetest."""
     logging.basicConfig(
         format="[{levelname}] {message}",
@@ -102,7 +116,18 @@ def main(ctx: click.Context, verbosity: int, quiet: bool, userpath: Optional[Pat
     if not userpath.exists():
         raise ApplicationError(f"Provided user path '{userpath}' doesn't exist")
 
-    ctx.obj = MinetestUserData(userpath=userpath)
+    cache = Path(appdirs.user_cache_dir("mtpkg", "twrightsman"))
+    if not cache.exists():
+        cache.mkdir()
+    if clear_cache:
+        logging.info("Clearing package metadata cache")
+        for child in cache.iterdir():
+            if child.is_dir():
+                shutil.rmtree(child)
+            else:
+                child.unlink()
+
+    ctx.obj = MinetestUserData(userpath=userpath, cache=cache)
 
 
 @main.command()
@@ -181,7 +206,7 @@ def install(
     # Validate package names
     valid_package_specs = validate_packages(package_str_set)
 
-    remote_db = ContentDB()
+    remote_db = ContentDB(cache=mtuserdata.cache)
     # Fetch latest minetest version if not provided
     if not mtver:
         mtver = remote_db.get_latest_minetest_version()
@@ -194,6 +219,8 @@ def install(
         remote_packages |= remote_db.get_recursive_dependencies(
             remote_packages, optional=install_optional, assume_yes=assume_yes
         )
+        # write cache after lots of requests in case of later crash
+        remote_db.write_cache()
 
     # Check for already-installed packages, remove those from remote_packages with a logging.info
     installable_packages = set()
@@ -253,6 +280,8 @@ def install(
                 f"'{game}' is not the game specified for world '{enable}' but it was defined as the base game for dependency resolution and mods were enabled for the world"
             )
 
+    remote_db.write_cache()
+
 
 @main.command()
 @click.option(
@@ -294,7 +323,7 @@ def update(
     else:
         valid_package_specs = validate_packages(package_str_set)
 
-    remote_db = ContentDB()
+    remote_db = ContentDB(cache=mtuserdata.cache)
     # Fetch latest minetest version if not provided
     if not mtver:
         mtver = remote_db.get_latest_minetest_version()
@@ -357,3 +386,5 @@ def update(
         with tempfile.TemporaryFile() as tmp:
             remote_db.download_release(package, latest_release_id, tmp)
             local_db.install_release(package, latest_release_id, tmp)
+
+    remote_db.write_cache()
