@@ -6,6 +6,7 @@ Module related to operations on packages installed to the Minetest user path
 
 
 from dataclasses import dataclass
+import io
 import logging
 from pathlib import Path
 import shutil
@@ -76,7 +77,10 @@ class LocalPackageDatabase:
                             f"'{mods_child}' has both a modpack.conf and a mod.conf"
                         )
                     # recursively add mods to mod2pack
-                    for submod in filter(lambda p: p.is_dir(), mods_child.iterdir()):
+                    for submod in filter(
+                        lambda p: p.is_dir() and (p / "mod.conf").exists(),
+                        mods_child.iterdir(),
+                    ):
                         if (submod / "mod.conf").exists():
                             submod_conf = parse_conf_file(submod / "mod.conf")
                             submod_name = submod_conf.get("name", submod.name)
@@ -259,7 +263,7 @@ class LocalPackageDatabase:
         world_mt_path = self.userpath / "worlds" / world_id / "world.mt"
         to_append = mods.copy()
         # check which mods are 1) enabled, 2) disabled, and 3) missing from world.mt
-        with tempfile.TemporaryFile("r+") as tmp:
+        with io.StringIO() as tmp:
             with open(world_mt_path) as world_mt_file:
                 for line in world_mt_file:
                     if line.startswith("load_mod_"):
@@ -284,7 +288,6 @@ class LocalPackageDatabase:
             for mod_name in to_append:
                 tmp.write(f"load_mod_{mod_name} = true\n")
             tmp.seek(0)
-            world_mt_path.unlink()
             with open(world_mt_path, "w") as world_mt_file:
                 shutil.copyfileobj(tmp, world_mt_file)
 
@@ -318,26 +321,23 @@ class LocalPackageDatabase:
             depth1_contents = list(ziproot.iterdir())
             conf_depth = None  # type: Optional[int]
             conf_filenames = get_possible_conf_names(package.type)
+            found_conf_filenames = set()
             if len(depth1_contents) == 1 and depth1_contents[0].is_dir():
                 # check if conf exists under dir, at depth 2
                 for child in depth1_contents[0].iterdir():
-                    if child.name in conf_filenames and conf_depth is None:
-                        conf_name = child.name
-                        conf_depth = 2
-                    elif child.name in conf_filenames and conf_depth is not None:
-                        raise LocalError(
-                            f"Multiple configuration files detected in package '{package.author}/{package.name}'. Aborting."
-                        )
+                    if child.name in conf_filenames:
+                        if conf_depth is None:
+                            conf_name = child.name
+                            conf_depth = 2
+                        found_conf_filenames.add(child.name)
             elif len(depth1_contents) > 1:
                 # check if conf exists at depth 1
                 for child in depth1_contents:
-                    if child.name in conf_filenames and conf_depth is None:
-                        conf_name = child.name
-                        conf_depth = 1
-                    elif child.name in conf_filenames and conf_depth is not None:
-                        raise LocalError(
-                            f"Multiple configuration files detected in package '{package.author}/{package.name}'. Aborting."
-                        )
+                    if child.name in conf_filenames:
+                        if conf_depth is None:
+                            conf_name = child.name
+                            conf_depth = 1
+                        found_conf_filenames.add(child.name)
             else:
                 raise LocalError(
                     f"Downloaded release {release_id} of '{package.author}/{package.name}' but it had no files in it!"
@@ -347,6 +347,12 @@ class LocalPackageDatabase:
                 raise LocalError(
                     f"Could not find an appropriate configuration file in release {release_id} of '{package.author}/{package.name}'!"
                 )
+            if ("modpack.conf" in found_conf_filenames) and (
+                "mod.conf" in found_conf_filenames
+            ):
+                raise LocalError(
+                    f"Found a modpack.conf and mod.conf at the top-level directory of '{package.author}/{package.name}'"
+                )
 
             # create folder to install to, removing existing one if necessary
             if install_path.exists():
@@ -374,9 +380,29 @@ class LocalPackageDatabase:
                 )
 
         # update the conf file with author and release ID
-        with open(install_path / conf_name, "a") as conf_file:
-            conf_file.write(f"author = {package.author}\n")
-            conf_file.write(f"release = {release_id}\n")
+        if conf_name == "modpack.txt":
+            conf_name = "modpack.conf"
+            logging.warning(
+                f"'{package.author}/{package.name}' has only a modpack.txt and no modpack.conf but did not specify a maximum Minetest version of 0.4.x, fixing locally"
+            )
+        conf_path = install_path / conf_name
+        if not conf_path.exists():
+            conf_path.touch()
+        with open(conf_path, "r") as conf_file:
+            new_conf_buf = io.StringIO()
+            for line in conf_file:
+                if line.split("=")[0].strip() == "name":
+                    continue
+                if not line.endswith("\n"):
+                    line += "\n"
+                new_conf_buf.write(line)
+            new_conf_buf.write(f"name = {package.name}\n")
+            new_conf_buf.write(f"author = {package.author}\n")
+            new_conf_buf.write(f"release = {release_id}\n")
+        with open(conf_path, "w") as conf_file:
+            new_conf_buf.seek(0)
+            shutil.copyfileobj(new_conf_buf, conf_file)
+        new_conf_buf.close()
 
         # add to package db
         local_pkg = LocalPackage(
@@ -389,18 +415,17 @@ class LocalPackageDatabase:
         self._pkgs[(package.type, package.name)] = local_pkg
 
         # recursively add mods to mod2pack
-        for submod in filter(lambda p: p.is_dir(), install_path.iterdir()):
-            if (submod / "mod.conf").exists():
-                submod_conf = parse_conf_file(submod / "mod.conf")
-                submod_name = submod_conf.get("name", submod.name)
-            else:
-                submod_name = submod.name
+        for submod in filter(
+            lambda p: p.is_dir() and (p / "mod.conf").exists(), install_path.iterdir()
+        ):
+            submod_conf = parse_conf_file(submod / "mod.conf")
+            submod_name = submod_conf.get("name", submod.name)
             self._mod2pack[submod_name] = local_pkg
 
 
 def get_possible_conf_names(pkgtype: PackageType) -> Set[str]:
     if pkgtype == PackageType.MOD:
-        return {"mod.conf", "modpack.conf"}
+        return {"mod.conf", "modpack.conf", "modpack.txt"}
     if pkgtype == PackageType.GAME:
         return {"game.conf"}
     if pkgtype == PackageType.TEXTURE_PACK:
@@ -411,8 +436,29 @@ def get_possible_conf_names(pkgtype: PackageType) -> Set[str]:
 def parse_conf_file(path: Path) -> Dict[str, str]:
     conf = {}
     with open(path) as conf_file:
-        for line in conf_file:
-            key, value = (e.strip() for e in line.split("=", maxsplit=1))
-            conf[key] = value
+        line = conf_file.readline()
+        while line:
+            line_split = line.split("=", maxsplit=1)
+            if len(line_split) != 2:
+                raise LocalError(
+                    f"Failed to parse line '{line.rstrip()}' in configuration file at {path}"
+                )
+            key = line_split[0].strip()
+            value = line_split[1].lstrip()
+            if value.startswith('"""'):
+                # multi-line value
+                value = value[3:]
+                if not value.rstrip().endswith('"""'):
+                    line = conf_file.readline()
+                    while line and (not line.rstrip().endswith('"""')):
+                        value += line
+                        line = conf_file.readline()
+                    if not line:
+                        raise LocalError(f"Couldn't parse configuration file at {path}")
+                    value += line.rstrip()[:-3]
+                else:
+                    value = value.rstrip()[:-3]
+            conf[key] = value.strip()
+            line = conf_file.readline()
 
     return conf
