@@ -11,14 +11,19 @@ from enum import Enum, unique
 from functools import partial
 import logging
 import operator
-from time import sleep
-from typing import Dict, Any, Set, FrozenSet, IO
+from pathlib import Path
+import pickle
+import time
+from typing import Dict, Any, Set, FrozenSet, IO, Union, List, Optional
 
 import click
 import requests
 from tqdm import tqdm  # type: ignore
 
 
+ParsedJSON = Union[List[Any], Dict[str, Any]]
+
+
 @unique
 class PackageType(Enum):
     MOD = 1
@@ -64,43 +69,78 @@ class RemotePackage:
 class ContentDB:
 
     _global_sleep: float = 0.4
-    _cache: Dict[str, Dict[str, requests.Response]]
+    _metadata_cache_name: str = "metadata.pkl"
+    _cache: Dict[str, Dict[str, Any]]
 
-    def __init__(self, server: str = "https://content.minetest.net"):
+    def __init__(
+        self, server: str = "https://content.minetest.net", cache: Optional[Path] = None
+    ):
         self.server = server
         self.base_url = f"{server}/api"
         self._cache = {"get": {}, "head": {}}
+        if cache is not None:
+            metadata_cache_path = cache / self._metadata_cache_name
+            # ignore cache if created more than 24 hours ago
+            if metadata_cache_path.exists():
+                if (time.time() - metadata_cache_path.stat().st_ctime) <= 86400:
+                    logging.debug("Loading persistent cache for ContentDB")
+                    with open(metadata_cache_path, "rb") as metadata_cache_file:
+                        self._cache = pickle.load(metadata_cache_file)
+                else:
+                    logging.debug("Cache too old, ignoring")
+        self.cache = cache
 
-    def _get_request(self, url: str, cache=True) -> requests.Response:
+    def write_cache(self):
+        if self.cache is not None:
+            metadata_cache_path = self.cache / self._metadata_cache_name
+            if metadata_cache_path.exists():
+                metadata_cache_path.unlink()
+            with open(metadata_cache_path, "wb") as metadata_cache_file:
+                pickle.dump(self._cache, metadata_cache_file)
+            logging.debug("Cache dumped successfully")
+
+    def _get_request_json(self, url: str) -> ParsedJSON:
         if url in self._cache["get"]:
-            response = self._cache["get"][url]
+            response_json = self._cache["get"][url]
+            if response_json == 404:
+                fake_response = requests.Response()
+                fake_response.status_code = 404
+                raise requests.HTTPError(response=fake_response)
         else:
             response = requests.get(url)
-            self._cache["get"][url] = response
-            sleep(self._global_sleep)
-        return response
+            if response.status_code == 404:
+                # cache 404 errors too
+                self._cache["get"][url] = 404
+            response.raise_for_status()
+            response_json = response.json()
+            self._cache["get"][url] = response_json
+            time.sleep(self._global_sleep)
+        return response_json
 
     def _head_request(
-        self, url: str, allow_redirects: bool = False, cache=True
-    ) -> requests.Response:
+        self, url: str, allow_redirects: bool = False
+    ) -> requests.structures.CaseInsensitiveDict:
         if url in self._cache["head"]:
-            response = self._cache["head"][url]
+            headers = self._cache["head"][url]
         else:
             response = requests.head(url, allow_redirects=allow_redirects)
-            sleep(self._global_sleep)
-        return response
+            response.raise_for_status()
+            headers = response.headers
+            self._cache["head"][url] = headers
+            time.sleep(self._global_sleep)
+        return headers
 
     def get_latest_minetest_version(self) -> str:
         """Fetches the latest stable Minetest version from the ContentDB.
 
         Returns the version name as a string and the integer client protocol version.
         """
-        response = self._get_request(f"{self.base_url}/minetest_versions/")
-        if response.status_code != 200:
+        try:
+            versions = self._get_request_json(f"{self.base_url}/minetest_versions/")
+        except requests.HTTPError as e:
             raise RemoteError(
-                f"Failed to get latest Minetest version from ContentDB. Server returned HTTP status code '{response.status_code}'."
-            )
-        versions = response.json()
+                "Failed to get latest Minetest version from ContentDB"
+            ) from e
         non_dev_versions = filter(lambda v: not v["is_dev"], versions)
         latest_version = sorted(
             non_dev_versions, key=lambda v: LooseVersion(v["name"]), reverse=True
@@ -109,16 +149,22 @@ class ContentDB:
         return latest_version["name"]
 
     def get_package(self, author: str, name: str) -> RemotePackage:
-        response = self._get_request(f"{self.base_url}/packages/{author}/{name}/")
-        if response.status_code != 200:
-            if response.status_code == 404:
+        try:
+            package_details = self._get_request_json(
+                f"{self.base_url}/packages/{author}/{name}/"
+            )
+        except requests.HTTPError as e:
+            if (e.response is not None) and e.response.status_code == 404:
                 raise PackageNotFoundError(
                     f"Package '{author}/{name}' not found in the ContentDB"
-                )
+                ) from e
             raise RemoteError(
-                f"Error retrieving package details for '{author}/{name}' from ContentDB. Server responded with status code '{response.status_code}'."
+                f"Error retrieving package details for '{author}/{name}' from ContentDB"
+            ) from e
+        if isinstance(package_details, list):
+            raise RemoteError(
+                f"Invalid JSON received while retrieving details for package '{author}/{name}'"
             )
-        package_details = response.json()
         return RemotePackage(
             name=package_details["name"],
             author=package_details["author"],
@@ -146,14 +192,18 @@ class ContentDB:
     def get_latest_compatible_release_id(
         self, package: RemotePackage, mtver: str, game_provided: bool
     ) -> int:
-        response = self._get_request(
-            f"{self.base_url}/packages/{package.author}/{package.name}/releases/"
-        )
-        if response.status_code != 200:
+        try:
+            releases = self._get_request_json(
+                f"{self.base_url}/packages/{package.author}/{package.name}/releases/"
+            )
+        except requests.HTTPError as e:
             raise RemoteError(
-                f"Error retrieving releases for '{package.author}/{package.name}' from ContentDB. Server responded with status code '{response.status_code}'"
+                f"Error retrieving releases for '{package.author}/{package.name}' from ContentDB"
+            ) from e
+        if isinstance(releases, dict):
+            raise RemoteError(
+                f"Invalid JSON received while retrieving releases for package '{package.author}/{package.name}'"
             )
-        releases = response.json()
         latest_release = releases[0]["id"]
 
         # filter releases to compatible and sort by descending release ID
@@ -213,12 +263,13 @@ class ContentDB:
         self, package: RemotePackage, release_id: int, dest: IO[bytes]
     ) -> None:
         download_url = self.get_release_download_url(package, release_id)
-        filesize_response = self._head_request(download_url, allow_redirects=True)
-        if filesize_response.status_code != 200:
+        try:
+            headers = self._head_request(download_url, allow_redirects=True)
+        except requests.HTTPError as e:
             raise RemoteError(
-                f"There was a problem requesting the file size of release '{release_id}' for package '{package.author}/{package.name}'. Server returned status code {filesize_response.status_code}."
-            )
-        filesize = int(filesize_response.headers["Content-Length"])
+                f"There was a problem requesting the file size of release '{release_id}' for package '{package.author}/{package.name}'"
+            ) from e
+        filesize = int(headers["Content-Length"])
         with requests.get(download_url, stream=True) as response, tqdm(
             desc=f"{package.author}/{package.name}",
             unit="B",
@@ -249,14 +300,19 @@ class ContentDB:
         # for each package, check if dependencies are in provided, otherwise add to required and queries
         while queries:
             query = next(iter(queries))
-            response = self._get_request(
-                query_url.format(author=query.author, name=query.name)
-            )
-            if response.status_code != 200:
+            try:
+                response_json = self._get_request_json(
+                    query_url.format(author=query.author, name=query.name)
+                )
+            except requests.HTTPError as e:
                 raise RemoteError(
                     f"Error retrieving dependencies for package '{query.author}/{query.name}'"
+                ) from e
+            if isinstance(response_json, list):
+                raise RemoteError(
+                    f"Invalid JSON received while retrieving dependencies for package '{query.author}/{query.name}'"
                 )
-            dependencies = response.json()[f"{query.author}/{query.name}"]
+            dependencies = response_json[f"{query.author}/{query.name}"]
             for dependency in dependencies:
                 if dependency["name"] not in (provided | required):
                     choices = []
